//
// MUZE KNOB FUNCTIONALITY
//
#include <Wire.h> // for I2C interface to color sensor.

#define COLORSENSOR 0x74

#define CAP_RED 0x06
#define CAP_GREEN 0x07
#define CAP_BLUE 0x08
#define CAP_CLEAR 0x09

#define INT_RED_LO 0x0A
#define INT_RED_HI 0x0B
#define INT_GREEN_LO 0x0C
#define INT_GREEN_HI 0x0D
#define INT_BLUE_LO 0x0E
#define INT_BLUE_HI 0x0F
#define INT_CLEAR_LO 0x10
#define INT_CLEAR_HI 0x11

#define DATA_RED_LO 0x40
#define DATA_RED_HI 0x41
#define DATA_GREEN_LO 0x42
#define DATA_GREEN_HI 0x43
#define DATA_BLUE_LO 0x44
#define DATA_BLUE_HI 0x45
#define DATA_CLEAR_LO 0x46
#define DATA_CLEAR_HI 0x47

#define OFFSET_RED 0x48
#define OFFSET_GREEN 0x49
#define OFFSET_BLUE 0x4A
#define OFFSET_CLEAR 0x4B

#define MAG_SENSE_MID_A     516
#define MAG_SENSE_MID_B     517
#define MAG_SENSE_MIN_DIFF  20
#define MAG_SENSE_MAX_DIFF  30
#define MAG_SENSE_MUL       60 / MAG_SENSE_MAX_DIFF

#define KNOB_DEGREES_PER_CLICK 22 

#define PIN_RGBSENSOR_LED   7
#define PIN_MAGSENSOR_A    14
#define PIN_MAGSENSOR_B    15

#define TOTAL_KNOBS 6
// Mapping of colors to knob mode...
const int KNOB_COLORS[TOTAL_KNOBS][4] = {
  {    
    656, 444, 131, 513                          } // ORANGE
  ,
  {
    754, 763, 517, 842                          } // WHITE
  ,
  {
    100, 116, 131, 142                          } // BLUE
  ,
  {
    95, 154, 116, 152                           } // GREEN
  ,
  {
    429, 203, 99, 285                           } // RED
  ,
  {
    175, 220, 298, 1000                         } // NOT ASSIGNED
  ,
};

#define MAX_COLOR_DIST 125
int colorSenseR = 0;
int colorSenseG = 0;
int colorSenseB = 0;
int colorSenseC = 0;
int closestKnobDist = MAX_COLOR_DIST;

// ############################################################################################
// Magnetic knob rotation sensing...
// ############################################################################################
void updateCurrentParameterBasedOnKnobRotation(){
  // senseKnobRotation(); // read knobAngle to determine if a knob is present. (not needed, we reed the knob in the note-play time of the code)...
  if(knobAngle == -1000){
    // NO KNOB...
    currentParameter = -1;
  }
  else{
    // Knob was present... get parameter by its color :)
    setCurrentParameterByKnobColor();
  }
}

void knobPlayRotationClickSound(){
  allowEvolve = 1;
  midiNoteOff(9, 75, 0); // wood block on 75/76
  midiNoteOn (9, 75, 80);
}

void knobPlayGotItSound(){
  midiNoteOff(9, 30, 0); // sticks on 30
  midiNoteOn (9, 30, 60);
}
void knobPlayLostItSound(){
  midiNoteOff(9, 31, 0); // sticks on 30
  midiNoteOn (9, 31, 60);
}

int getAngleDelta(int angleA, int angleB){
  int delta = angleA - angleB;
  if(delta > 180){
    delta = delta - 360;
  }
  if(delta < -180){
    delta = delta + 360;
  }
  return delta;
}

boolean sensedKnobLastTime = false;
void senseKnobRotationAndColor(){
  if(sensedKnobLastTime && knobAngle != -1000){
    performColorMeasurement();
    sensedKnobLastTime = false;
  }else{
    senseKnobRotation();
    sensedKnobLastTime = true;
  }
}

void senseKnobRotation(){
  // this function senses two orthogonal magnetic fields and determines the magnet's rotation.
  // after this has been run, other code can check:
  //   knobAngle == -1000 (no knob was seen)
  //   
  magSenseA = max(-MAG_SENSE_MAX_DIFF, min(MAG_SENSE_MAX_DIFF, analogRead(PIN_MAGSENSOR_A) - MAG_SENSE_MID_A));
  magSenseB = max(-MAG_SENSE_MAX_DIFF, min(MAG_SENSE_MAX_DIFF, analogRead(PIN_MAGSENSOR_B) - MAG_SENSE_MID_B));

  knobAngle = -1000; // assume no knob is present...
  if(abs(magSenseA) > abs(magSenseB)){
    if(abs(magSenseA) > MAG_SENSE_MIN_DIFF){
      // find the position, |A| > |B|
      if(magSenseA > 0){
        knobAngle =  45 + magSenseB*MAG_SENSE_MUL;
      }
      else{
        knobAngle = 225 - magSenseB*MAG_SENSE_MUL;
      }
    }
  }
  else{
    if(abs(magSenseB) > MAG_SENSE_MIN_DIFF){
      // find the position, |B| > |A|
      if(magSenseB > 0){
        knobAngle = 135 - magSenseA*MAG_SENSE_MUL;
      }
      else{
        knobAngle = 315 + magSenseA*MAG_SENSE_MUL;
      }
    }
  }
  // make sure the angle is clean (max/min as well as CW vs CCW).
  if(knobAngle > -100){
    knobAngle = max(0, min(359, 360-knobAngle));
    if(lastKnobAngle == -1){
      // this is the first time we're seing the knob!
      lastKnobAngle = knobAngle;
      knobPlayGotItSound();
    }
    knobDiff  = getAngleDelta(knobAngle, lastKnobAngle);
    if(abs(knobDiff) >= KNOB_DEGREES_PER_CLICK){
      // the knob was turned -- send it!
      encoderSend += knobDiff / KNOB_DEGREES_PER_CLICK;
      lastKnobAngle = knobAngle;
      knobPlayRotationClickSound();
      //hapticClick();
    }
  }
  else{
    if(lastKnobAngle != -1){
      // knob was just removed...
      lastKnobAngle = -1;
      knobPlayLostItSound();
    }
  } 
}

// ############################################################################################
// RGB Color Sensor functions
// derived from: http://interactive-matter.org/2008/08/tinkering-with-colorSensor-s371-q999/
// ############################################################################################
void setCurrentParameterByKnobColor(){
  int closestIndex = -1;
  int dist = 0;
  int i = 0;
  closestKnobDist = 4096;
  //performColorMeasurement();   // get color info!
  for(i=0; i<TOTAL_KNOBS; i++){
    dist  = abs(colorSenseR - KNOB_COLORS[i][0]);
    dist += abs(colorSenseG - KNOB_COLORS[i][1]);
    dist += abs(colorSenseB - KNOB_COLORS[i][2]);
    dist += abs(colorSenseC - KNOB_COLORS[i][3]);
    if(dist < closestKnobDist){
      closestKnobDist = dist;
      if(dist < MAX_COLOR_DIST){
        closestIndex = i;
      }
    }
  }
  currentParameter = closestIndex;
  if(DEBUG){
    Serial.print("RGBC = (");
    Serial.print(colorSenseR);
    Serial.print(", ");
    Serial.print(colorSenseG);
    Serial.print(", ");
    Serial.print(colorSenseB);
    Serial.print(", ");
    Serial.print(colorSenseC);
    Serial.print("),closestKnobDist = ");
    Serial.print(closestKnobDist);
    Serial.print(", knobAngle = ");
    Serial.println(knobAngle);
  }
}
static void colorSensor_init(){
  pinMode(PIN_RGBSENSOR_LED, OUTPUT);  // declare the PIN_RGBSENSOR_LED as an OUTPUT
  // first, start up SPI communication interface.
  Wire.begin();
  // sensor gain setting (Avago app note 5330)
  // CAPs are 4bit (higher value will result in lower output)
  write_register(CAP_RED,   0x01);
  write_register(CAP_GREEN, 0x01);
  write_register(CAP_BLUE,  0x01);
  write_register(CAP_CLEAR, 0x01);
  // Gain is between 1 and 4096 
  set_gain(INT_RED_LO,   1200);
  set_gain(INT_GREEN_LO, 1200);
  set_gain(INT_BLUE_LO,  1200);
  set_gain(INT_CLEAR_LO,  600);
}
static void write_register(uint8_t register_name, uint8_t register_value){
  Wire.beginTransmission(COLORSENSOR);
  Wire.send(register_name);
  Wire.send(register_value);
  Wire.endTransmission();
}
static uint8_t read_register(uint8_t register_name){
  Wire.beginTransmission(COLORSENSOR);
  Wire.send(register_name);
  Wire.endTransmission();
  Wire.requestFrom(COLORSENSOR,2);
  while (Wire.available() < 1) {
    // waiting...
  }
  uint8_t result = Wire.receive();
  return result;
}
void set_gain(int gainRegister, int gain) {
  if (gain <4096) {
    uint8_t hi = gain >> 8;
    uint8_t lo = gain;
    write_register(gainRegister, lo);
    write_register(gainRegister+1, hi);
  }
}
// this is a helper that does two reads.
int get_readout(int readRegister) {
  return read_register(readRegister) + (read_register(readRegister+1) << 8);
}
void performColorMeasurement() {
  digitalWrite(PIN_RGBSENSOR_LED,HIGH);
  delay(1);
  write_register(0x00,0x01); // start sensing
  while(read_register(0x00) != 0) {
    // waiting for a result
  }
  colorSenseR = get_readout(DATA_RED_LO);
  colorSenseG = get_readout(DATA_GREEN_LO);
  colorSenseB = get_readout(DATA_BLUE_LO);
  colorSenseC = get_readout(DATA_CLEAR_LO);
  digitalWrite(PIN_RGBSENSOR_LED,LOW);
}



