
#define DEBUG false

const int MUZE_VERSION = 81;

#define VOICES_MAX 8  // maximum number of voices
// and then, these all need to be in sync with the number of voices..
int phraseLengths[VOICES_MAX] = { 
  32, 16, 8, 16, 8, 4, 64, 16    };
int noteDurations[VOICES_MAX] = {
  2, 2, 4, 8, 2, 8, 2, 2       };
int rootNotes    [VOICES_MAX] = {
  48, 48, 36, 36, 48, 48, 48, 36 };
int scaleModes   [VOICES_MAX] = {
  0, 0, 0, 0, 0, 0, 0, 0      };
int voicesPlaying[VOICES_MAX] = {
  1, 0, 0, 0, 0, 0, 0, 0      };

#define PHRASE_LENGTH_MAX    64 // maximum length of sequence phrase
#define PHRASE_LENGTH_MIN    4
#define DENSITY_MIN          16
#define STACCATO_MAX         80
#define STACCATO_MIN         40
#define TEMPO_MIN            125
#define TEMPO_MAX            200
#define DURATION_MIN         1
#define DURATION_MAX         8

#define MIDI_MIN_NOTE_ON_VELOCITY 70
#define MIDI_NOTE_ON_RANGE        30

int pressState       = 0;
int currentParameter = 0; // to cycle between modes
int currentMood      = 0;
int currentDensity   = 8;
int tempo            = 150;
int staccato         = 70;

int allowEvolve      = 0;

volatile int encoderSend = 0; // shuttles encoder +/- value to current parameter

#define NOTES_IN_MOOD        16
#define TOTAL_MOODS          3
int moods[TOTAL_MOODS][NOTES_IN_MOOD] = {
  { // Happy
    3, 5, 7, 8, 10, 12, 15, 17, 19, 20, 22, 24, 27, 29, 31, 32           }
  ,{ // Sad
    0, 2, 5, 7, 10, 12, 14, 17, 19, 22, 24, 26, 29, 31, 34, 36           }
  //,{ // Romantic
  //  0, 2, 3, 5, 7, 10, 12, 14, 15, 17, 19, 22, 24, 26, 27, 29            } 
  ,{ // Quirky
    0, 2, 4, 7, 9, 12, 14, 16, 19, 21, 24, 26, 28, 31, 33, 36            }
  /*,{ // Major
    0, 2, 4, 5, 7, 9, 11, 12, 14, 16, 17, 19, 21, 23, 24, 26             }
  */
};
int *mood; // makes an alias/pointer array for the scale selection.

// array for melody construction...
int sequence[VOICES_MAX][PHRASE_LENGTH_MAX]; // note-pointers (offset in the current mood) to play at each time slot.
int velocity[VOICES_MAX][PHRASE_LENGTH_MAX]; // array for the composition midi velocities

// Magnetic Knob Settings...
int magSenseA     =  0; 
int magSenseB     =  0;
int knobAngle     =  0;
int lastKnobAngle = -1;
int knobDiff      =  0;

// ############################################################################################
// SETUP
// ############################################################################################
void setup()
{
  randomSeed(analogRead(0)+analogRead(1)+analogRead(2)+analogRead(3)+analogRead(4));

  // init serial communication
  Serial.begin(115200); 
  Serial.print("Muze 0.0.");
  Serial.print(MUZE_VERSION/10);
  Serial.print(".");
  Serial.println(MUZE_VERSION%10);

  // load initial sequence and thencombine sequence and mood to make melody
  int timeToLoad = loadSavedState();
  Serial.print("Complete.  Loaded in ");
  Serial.print(timeToLoad);
  Serial.println("msec.");

  // Color sensor setup...
  colorSensor_init();

  // MIDI setup...
  initMIDI();
  
  // for now, force a load of BicycleBuiltForTwo. :)
  //loadBicycleBuiltForTwo();
}

// ############################################################################################
// MAIN LOOP :)
// ############################################################################################
long loopNumber = 0L;
long startBeatTime = 0L;
void loop(){
  loopNumber++;
  // iterate over the notes of the melody:
  Serial.print("Loop: "); 
  Serial.println(loopNumber);
  mood = moods[currentMood];
  for (int p = 0; p < PHRASE_LENGTH_MAX; p++) {
    startBeatTime - millis();
    
    // play the compostion notes already!
    for (int v = 0; v < VOICES_MAX ; v++) {
      int pD = p / noteDurations[v];
      if(velocity[v][pD] > 0 && (p % noteDurations[v]) == 0 && voicesPlaying[v]==1){
        muzePlayNote(getMIDINoteAtPosition(v, pD), velocity[v][pD]);
      }
    }
    float fStaccato = (float)staccato/100.0;

    // delay while the note is playing...
    int timeSpentSaving = saveStateStep();
    inputSafeDelay(1+(int)(fStaccato*tempo)-timeSpentSaving);
    for (int v = 0; v < VOICES_MAX ; v++) {
      int pD = p / noteDurations[v];
      if((p % noteDurations[v]) == (noteDurations[v]-1) && voicesPlaying[v]==1){
        muzeStopNote(getMIDINoteAtPosition(v, pD));
      }
    }

    // listen for incomming midi data and print it, not currently necessary (but for maybe later).
    //checkForMIDIData(); 
    // handle knob input to update variables.
    updateCurrentParameterBasedOnKnobRotation();
    mysteriouslyUpdateVariablesBasedOnCurrentParamAndKnob();
    // make any last-minute changes to the sequence...
    updateSequence();
    // change mood if needed (but only when we're sure all notes are off)...
    if(p % DURATION_MAX == (DURATION_MAX-1)){
      mood = moods[currentMood]; // only change the mood after a complete phrase (or we risk stuck-on notes!).
    }

    // time with notes complete...
    inputSafeWaitTillMillis(startBeatTime + tempo);
    //inputSafeDelay(tempo - (int)(fStaccato*tempo));
  }
  if(DEBUG){
    int newProg = loopNumber % 128;
    midiSetProgram(1, newProg);
    Serial.print("New MIDI Prog: ");
    Serial.println(newProg, DEC);
  }
  // EVOLUTION!
  muzeAIEvolve();
}


// ############################################################################################
// ############################################################################################
// this is a delay routine that performs other tasks while waiting (so we don't completely clog up the works)!
// NOTE: Every 50 days, there is a chance this will overflow and produce a bad delay (but should just return prematurely).
unsigned long startTime = 0UL;
void inputSafeDelay(int sleepTime){
  if(sleepTime < 0){
    return;
  }
  if(sleepTime < 10){
    delay(sleepTime);
  }
  else{
    startTime = millis();
    inputSafeWaitTillMillis(startTime + sleepTime);
  }
}
void inputSafeWaitTillMillis(long endMillis){
  while(millis() < endMillis){
      senseKnobRotationAndColor();
      //senseKnobRotation();
    }
}
















